/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_connection.cc
 *
 *   DESCRIPTION
 *      implementation of the Pg::Connection class.
 *   Pg::Connection encapsulates a frontend to backend connection
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#include <stdexcept>

#include "pg_connection.h"
#include "pg_result.h"


using namespace std;


Pg::Connection::Connection(const string &ConnInfo) :
  m_ConnInfo(ConnInfo),
  m_Conn(0)
{
  Connect();
}


Pg::Connection::~Connection()
{
  Disconnect();
}


void Pg::Connection::Connect()
{
  Disconnect();
  m_Conn = PQconnectdb(m_ConnInfo.c_str());

  if (!m_Conn)
    throw broken_connection();

  if (!IsOpen())
  {
    const string Msg( ErrMsg() );
    Disconnect();
    throw broken_connection(Msg);
  }

  if (Status() != CONNECTION_OK)
  {
    const string Msg( ErrMsg() );
    Disconnect();
    throw runtime_error(Msg);
  }
}


void Pg::Connection::Disconnect() throw ()
{
  if (m_Conn)
  {
    PQfinish(m_Conn);
    m_Conn = 0;
  }
}


bool Pg::Connection::IsOpen() const
{
  return m_Conn && (Status() != CONNECTION_BAD);
}


bool Pg::Connection::GetNotify(string *Name, int *be_pid)
{
  PGnotify *N = PQnotifies(m_Conn);
  if (!N)
  {
    // None in queue locally, but there notification may be one pending at the
    // back-end.  Ask for an update and retry.
    // This should theoretically be faster than calling PQconsumeInput() every
    // time as the libpq manual suggests.
    PQconsumeInput(m_Conn);
    N = PQnotifies(m_Conn);
  }

  if (!N) return false;

  try
  {
    if (Name) *Name = N->relname;
    if (be_pid) *be_pid = N->be_pid;
  }
  catch (...)
  {
    free(N);
    throw;
  }

  free(N);

  return true;
}


int Pg::Connection::Status() const
{
  return PQstatus(m_Conn);
}


const char *Pg::Connection::ErrMsg() const
{
  return m_Conn ? PQerrorMessage(m_Conn) : "No connection to database";
}


Pg::Result Pg::Connection::Exec(const char Q[], 
                                int Retries, 
				const char OnReconnect[])
{
  if (!m_Conn)
    throw runtime_error("No connection to database");

  Result R( PQexec(m_Conn, Q) );

  while ((Retries > 0) && !R.c_ptr() && !IsOpen())
  {
    Retries--;

    PQreset(m_Conn);

    if (OnReconnect)
    {
      Result Temp( PQexec(m_Conn, OnReconnect) );
      if (!Temp.c_ptr()) continue;

      Temp.CheckStatus();
    }

    R = PQexec(m_Conn, Q);
  }

  if (!R.c_ptr()) throw broken_connection();
  R.CheckStatus();

  return R;
}

