/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_connection.h
 *
 *   DESCRIPTION
 *      implementation of the Pg::Connection class.
 *   Pg::Connection encapsulates a frontend to backend connection
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_CONNECTION_H
#define PG_CONNECTION_H

#include <stdexcept>

#include "pg_util.h"
#include "pg_transactor.h"


// TODO: Support triggers
// TODO: Disallow multiple simultaneous transactions on same connection


namespace Pg
{
class Result;	// See pg_result.h


// Exception class for lost backend connection
// (May be changed once I find a standard exception class for this)
class broken_connection : public runtime_error
{
public:
  broken_connection() : runtime_error("Connection to back end failed") {}
  explicit broken_connection(const string &whatarg) : runtime_error(whatarg) {}
};


// Connection class.  This is the first class to look at when you wish to work
// with a database through libpqxx.  It is automatically opened by its
// constructor, and automatically closed upon destruction, if it hasn't already
// been closed manually.
// To query or manipulate the database once connected, use the Transaction
// class (see pg_transaction.h) or preferably the Transactor framework (see 
// pg_transactor.h).
class Connection
{
public:
  explicit Connection(const std::string &ConnInfo);
  ~Connection();

  void Disconnect() throw ();

  bool IsOpen() const;

  template<typename TRANSACTOR> void Perform(const TRANSACTOR &, int Attempts=3);

  // Get notification from queue, returning whether one was pending.  
  // Information about the message is filled into variables pointed to by the
  // arguments, insofar as they are non-null.
  bool GetNotify(std::string *Name=0, int *be_pid=0);

private:
  void Connect();
  int Status() const;
  const char *ErrMsg() const;

  std::string m_ConnInfo;
  PGconn *m_Conn;

  friend class Transaction;
  Result Exec(const char[], int Retries=3, const char OnReconnect[]=0);

  // Not allowed:
  Connection(const Connection &);
  Connection &operator=(const Connection &);
};



// Invoke a Transactor, making at most Attempts attempts to perform the
// encapsulated code on the database.  If the code throws any exception other
// than broken_connection, it will be aborted right away.
// Take care: neither abort() nor commit() will be invoked on the original
// transactor you pass into the function.  It only serves as a prototype for
// the transaction to be performed.  In fact, this function may copy-construct
// any number of Transactors from the one you passed in, calling either 
// commit() or abort() only on those that actually have their operator()
// invoked.
template<typename TRANSACTOR> 
inline void Connection::Perform(const TRANSACTOR &T,
                                int Attempts)
{
  Done = (Attempts > 0);

  // Make attempts to perform T
  while (!Done)
  {
    --Attempts;

    // Work on a copy of T2 so we can restore the starting situation if need be
    TRANSACTOR T2(T);
    try
    {
      Transaction X(*this);
      T2(X);
    }
    catch (const broken_connection &)
    {
      // Clear-cut case: retry until attempts count runs out
      T2.abort();
      if (Attempts <= 0) throw;
      continue;
    }
    catch (...)
    {
      // Could be any kind of error.  If connection has gone bad, assume that
      // may be the cause of the problem and retry until attempts count runs
      // out
      T2.abort();
      if ((Attempts <= 0) || IsOpen()) throw;
      continue;
    }
    T2.commit();
    Done = true;
  }

}


}

#endif

