/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_result.h
 *
 *   DESCRIPTION
 *      definitions for the Pg::Result class and support classes.
 *   Pg::Result represents the set of result tuples from a database query
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_RESULT_H
#define PG_RESULT_H

#include <iterator>
#include <stdexcept>
#include <typeinfo>

#include "pg_util.h"

// TODO: Handle binary tuples
// TODO: Async data transfer
// TODO: Cursors as forward-only iterators

namespace Pg
{

typedef long Result_size_type;
typedef int Tuple_size_type;


// Query or command result set.  This behaves pretty much as a container (as 
// defined by the C++ standard library) and provides random access const
// iterators to iterate over its tuples.  A tuple can also be accessed by
// indexing a Result R by the tuple's zero-based number:
//
//	for (Result::size_type i=0; i < R.size(); ++i) Process(R[i]);
//
class Result
{
public:
  Result();
  Result(const Result &);
  ~Result();
  
  Result &operator=(const Result &);

  typedef Result_size_type size_type;
  class Field;

  // TODO: Define const_reference typedef
  
  // A Tuple is really a reference to one entry in a Result.  It also acts as a
  // container mapping column numbers or names to Field values (see below):
  //
  // 	cout << Tuple["date"].c_str() << ": " << Tuple["name"].c_str() << endl;
  //
  // The fields in a Tuple can not currently be iterated over.
  class Tuple
  {
  public:
    typedef Tuple_size_type size_type;

    Tuple() {}
    Tuple(const Result *r, size_type i) : m_Home(r), m_Index(i) {}
    ~Tuple() {} // Yes Scott Meyers, you're absolutely right (1)

    Field operator[](size_type i) const { return Field(*this, i); }
    Field operator[](const char[]) const;
    Field operator[](std::string s) const {return operator[](s.c_str());}
    Field at(size_type) const;
    Field at(const char[]) const;
    Field at(std::string s) const { return at(s.c_str()); }

    size_type size() const;

    Result::size_type Row() const { return m_Index; }

  protected:
    const PGresult *Result_c_ptr() const { return m_Home->c_ptr(); }
    const Result *m_Home;
    Result::size_type m_Index;
  };


  // A Field is one entry in a Tuple.  It represents an actual value in the
  // result set, and can be converted to various types.
  class Field : private Tuple
  {
  public:
    // TODO: Retrieve field type using PQftype()?
    // TODO: Support PQfmod()
    Field(const Tuple &Row, Tuple::size_type Col) : Tuple(Row), m_Col(Col) {}

    const char *c_str() const;	// Read as plain C string
    const char *name() const;	// Column name

    // Read value into Obj; or leave Obj untouched & return false if null
    template<typename T> bool to(T &Obj) const
    {
      if (is_null())
	return false;
      const char *val = c_str();

      if (!sscanf(val, FmtString(Obj), &Obj))
	throw std::runtime_error("Cannot convert field " + 
			         std::string(name()) +
			         " from \"" + val + "\" "
				 "to " + typeid(T).name());
      return true;
    }

    // Read value into Obj; or use Default & return false if null
    template<typename T> bool to(T &Obj, const T &Default) const
    {
      const bool NotNull = to(Obj);
      if (!NotNull)
	Obj = Default;
      return NotNull;
    }

    bool is_null() const;

    int size() const;

  private:

    Tuple::size_type m_Col;
  };


  // A Result, once obtained, cannot be modified.  However its const_iterator
  // type can be used to inspect its Tuples without changing them.
  class const_iterator : 
    public std::iterator<random_access_iterator_tag, 
                         const Tuple,
                         Result::size_type>, 
    public Tuple
  {
  public:

    // Confusing but necessary (I think, but I'm ill right now, besides living
    // in the bad parts of Amsterdam--so plenty of reasons why I might be
    // hallucinating).  The iterator "points to" its own Tuple, which is also
    // itself.  This allows a Result to be addressed as a two-dimensional
    // container without going through the intermediate step of dereferencing
    // the iterator.  I hope this works out to be similar to C pointer/array
    // semantics in useful cases (2).
    pointer operator->()  const { return this; }
    reference operator*() const { return *operator->(); }

    const_iterator operator++(int);
    const_iterator &operator++();
    const_iterator operator--(int);
    const_iterator &operator--();

    const_iterator &operator+=(difference_type);
    const_iterator &operator-=(difference_type);

    bool operator==(const const_iterator &) const;
    bool operator!=(const const_iterator &) const;
    bool operator<(const const_iterator &) const;
    bool operator<=(const const_iterator &) const;
    bool operator>(const const_iterator &) const;
    bool operator>=(const const_iterator &) const;

    const_iterator operator+(difference_type o) const
    {
      return const_iterator(m_Home, m_Index + o);
    }
    friend const_iterator operator+(difference_type o, const_iterator i)
    {
      return i + o;
    }

    const_iterator operator-(difference_type o) const
    {
      return const_iterator(m_Home, m_Index - o);
    }

    difference_type operator-(const_iterator i) const { return num()-i.num(); }

     Result::size_type num() const { return Row(); }

  private:
    friend class Result;
    const_iterator(const Result *r, Result::size_type i) : Tuple(r, i) {}
  };

  const_iterator begin() const { return const_iterator(this, 0); }
  const_iterator end() const;
  // TODO: Handle reverse iterators
  // TODO: front()?

  size_type size() const;
  bool empty() const;

  const Tuple operator[](size_type i) const { return Tuple(this, i); }
  const Tuple at(size_type i) const;


private:
  friend class Result::Tuple;
  const PGresult *c_ptr() const { return m_Result; }

  PGresult *m_Result;
  mutable int *m_Refcount;

  friend class Connection;
  explicit Result(PGresult *);
  Result &operator=(PGresult *);
  void CheckStatus() const;


  void MakeRef(PGresult *);
  void MakeRef(const Result &);
  void LoseRef() throw ();
};




template<> inline bool Result::Field::to(std::string &Obj) const
{
  if (is_null())
    return false;
  Obj = std::string(c_str());
  return true;
}

template<> inline bool Result::Field::to(const char *&Obj) const
{
  if (is_null()) 
    return false;
  Obj = c_str();
  return true;
}


} // namespace


/* 
(1) Scott Meyers, in one of his essential books, "Effective C++" and "More 
Effective C++", points out that it is good style to have any class containing 
a member of pointer type define its own destructor--just to show that it knows
what it is doing.  This helps prevent nasty memory leak / double deletion bugs
typically resulting from programmers' omission to deal with such issues in
their destructors.

The -Weffc++ option in gcc generates warnings for noncompliance with Scott's
style guidelines, and hence necessitates the definition of this destructor,\
trivial as it may be.

(2) IIRC Alex Stepanov, the inventor of the STL, once remarked that having
this as standard behaviour for pointers would be useful in some algorithms.
So even if this makes me look foolish, I would seem to bein distinguished 
company.
*/

#endif

