/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_transaction.cc
 *
 *   DESCRIPTION
 *      implementation of the Pg::Transaction class.
 *   Pg::Transaction represents a database transaction
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#include <stdexcept>

#include "pg_connection.h"
#include "pg_result.h"
#include "pg_transaction.h"


using namespace std;


Pg::Transaction::Transaction(Connection &Conn) :
  m_Conn(Conn),
  m_Status(st_nascent)
{
  Begin();
}


Pg::Transaction::~Transaction()
{
  if (m_Status != st_committed)
    Abort();
}


void Pg::Transaction::Commit()
{
  // Check previous status code.  Caller should only call this function if
  // we're in "implicit" state (neither committed nor explicitly aborted).
  switch (m_Status)
  {
  case st_nascent:	// Probably can't happen.  Skip commit entirely.
    m_Status = st_aborted;
    return;

  case st_clean:	// Weird, but no skin off our nose.
  case st_implicit:	// Just fine.  This is what we expect.
    break;

  case st_aborted:
    throw logic_error("Attempt to commit previously aborted database transaction");

  case st_committed:
    throw logic_error("Database transaction committed multiple times");

  default:
    throw logic_error("Internal libpqxx error: Pg::Transaction: invalid status code");
  }

  try
  {
    m_Conn.Exec("COMMIT WORK", 0);
  }
  catch (...)
  {
    m_Status = st_aborted;
    throw;
  }

  m_Status = st_committed;
}


void Pg::Transaction::Abort()
{
  // Check previous status code.  Quietly accept multiple aborts to 
  // simplify emergency bailout code.
  switch (m_Status)
  {
  case st_nascent:	// Never began transaction.  No need to issue rollback.
    break;

  case st_clean:
  case st_implicit:
  case st_aborted:
    try
    {
      m_Conn.Exec("ABORT WORK", 0);
    }
    catch (...)
    {
    }
    break;

  case st_committed:
    throw logic_error("Attempt to abort previously committed transaction");

  default:
    throw logic_error("Internal libpqxx error: Pg::Transaction: invalid status code");
  }

  m_Status = st_aborted;
}


Pg::Result Pg::Transaction::Exec(const char C[])
{
  int Retries;

  switch (m_Status)
  {
  case st_nascent:
    Begin();
    Retries = 2;
    break;

  case st_clean:
    Retries = 3;
    break;

  case st_implicit:
    Retries = 0;
    break;

  case st_committed:
    throw logic_error("Attempt to execute query in committed transaction");

  case st_aborted:
    throw logic_error("Attempt to execute query in aborted transaction");

  default:
    throw logic_error("Internal libpqxx error: Pg::Transaction: invalid status code");
  }

  Result R;
  try
  {
    R = m_Conn.Exec(C, Retries, "BEGIN WORK");
    NoteCmd();
  }
  catch (...)
  {
    Abort();
    throw;
  }

  return R;
}



void Pg::Transaction::Begin()
{
  m_Conn.Exec("BEGIN WORK");
  m_Status = st_clean;
}


void Pg::Transaction::NoteCmd()
{
  if (m_Status == st_clean)
    m_Status = st_implicit;
}

