/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_transaction.h
 *
 *   DESCRIPTION
 *      definition of the Pg::Transaction class.
 *   Pg::Transaction represents a database transaction
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_TRANSACTION_H
#define PG_TRANSACTION_H

/* While you may choose to interface to the database directly through a
 * Transaction object, it is recommended that you wrap your transaction
 * code into a Transactor code instead and feed that to the Transaction.  
 * See pg_transactor.h for more about Transactor.
 *
 * If using a Transactor makes your code less portable or too complex, go 
 * ahead, use a Transaction anyway.
 */

#include "pg_util.h"


namespace Pg
{
class Connection; // See pg_connection.h
class Result; // See pg_result.h

// Use a Transaction object to enclose operations on a database in a single
// "unit of work."  This ensures that the whole series of operations either
// succeeds as a whole or fails completely.  In no case will it leave half
// finished work behind in the database.
// Queries can currently only be issued through a Transaction to encourage good
// practice.  If this is ever changed, that is likely to be accomplished by
// adding an "unprotected" variant of the Transaction class to maintain 
// symmetry and facilitate introduction of transactional integrity into client
// code.
// Once processing on a transaction has succeeded and any changes should be
// allowed to become permanent in the database, call Commit().  If something
// has gone wrong and the changes should be forgotten, call Abort() instead.
// If you do neither, an implicit Abort() is executed at destruction time.
// It is an error to either commit or abort a Transaction that has already
// been committed, or to commit a transaction that has already been aborted.
// Aborting an already aborted transaction has been allowed to facilitate error
// handling code.  Repeated aborts on the same transaction have the same effect
// as a single abort.
class Transaction
{
public:
  explicit Transaction(Connection &);
  ~Transaction();

  void Commit();
  void Abort();

  // Execute query directly
  Result Exec(const char[]);

private:
  /* A Transaction goes through the following stages in its lifecycle:
   *  - nascent: the transaction hasn't actually begun yet.  If our
   *    connection fails at this stage, the Connection may recover and 
   *    the Transaction can attempt to establish itself again.
   *  - clean: the transaction has begun, but no action has yet been
   *    taken inside it.  If the connection fails, it can be reinstated
   *    safely by starting a new backend transaction once the connection
   *    has recovered.
   *  - implicit: the transaction is in full progress.  Since no commit
   *    command has been issued, abortion is implicit if the connection
   *    fails now.
   *  - aborted: an abort has been issued; the transaction is terminated
   *    and its changes to the database rolled back.  It will accept no 
   *    further commands.
   *  - committed: the transaction has completed successfully, meaning
   *    that a commit has been issued.  No further commands are accepted.
   */
  enum Status { st_nascent, st_clean, st_implicit, st_aborted, st_committed };

  void Begin();

  // Update status code to reflect execution of a command or query
  void NoteCmd();

  Connection &m_Conn;

  Status m_Status;

  // Not allowed:
  Transaction();
  Transaction(const Transaction &);
  Transaction &operator=(const Transaction &);
};

}

#endif

