/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_transactor.h
 *
 *   DESCRIPTION
 *      definition of the Pg::Transactor class.
 *   Pg::Transactor is a framework-style wrapper for safe transactions
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */

/* Some transactions may be replayed if their connection fails, until they do 
 * succeed.  These can be encapsulated in Transactor-derived classes.  The 
 * Transactor will take care the aborting and retrying if its connection goes 
 * bad.
 *
 * The Transactor wrapper also provides a framework for doing this safely,
 * avoiding typical pitfalls and encouraging programmers to separate their
 * transaction definitions (essentially, business rules implementations) from 
 * their higher-level code (application using those business rules).
 *
 * Pass an object of your Transactor-based class to Connection::Perform() to
 * perform its actions.  
 *
 * Connection::Perform() is actually a template, specializing itself to any
 * Transactor type you pass to it.  This means you will have to pass it a 
 * reference of your object's ultimate static type; runtime polymorphism is
 * not allowed.  Hence the absence of virtual methods in Transactor.  The
 * exact methods to be called at runtime *must* be resolved at compile time.
 *
 * Note that your Transactor-derived class must define a copy constructor.
 * This will be used to create a "clean" copy of your transactor for every
 * attempt that Perform() makes to run it.  See the definition of Perform()
 * below.
 */

namespace Pg
{
class Transaction;

class Transactor
{
public:
  // Overridable transaction definition.  Will be retried if connection goes
  // bad, but not if some other exception is thrown.
  void operator()(Transaction &);

  // Overridable member functions, called by Connection::Perform() if attempt 
  // to run transaction fails/succeeds, respectively.  If operator() is never
  // invoked, neither abort() nor commit() will be called.  Your code must be
  // able to cope with this.
  void abort() {}
  void commit() {}
};

}

