/*-------------------------------------------------------------------------
 *
 *   FILE
 *	pg_util.h
 *
 *   DESCRIPTION
 *      Various utility definitions for libpqxx
 *
 * Copyright (c) 2001, Jeroen T. Vermeulen <jtv@xs4all.nl>
 *
 *-------------------------------------------------------------------------
 */
#ifndef PG_UTIL_H
#define PG_UTIL_H

#include <cstdio>
#include <string>

extern "C"
{
#include "postgres_fe.h"
#include "libpq-fe.h"
}

namespace Pg
{
// C-style format strings for various built-in types.  Only allowed for
// certain types, for which this function is explicitly specialized below.
template<typename T> inline const char *FmtString(const T &);

// Not implemented to prevent accidents with irregular meaning of argument:
// template<> inline const char *FmtString(const char *const &) { return "%s"; }

template<> inline const char *FmtString(const int &)           { return  "%i"; }
template<> inline const char *FmtString(const long &)          { return "%li"; }
template<> inline const char *FmtString(const unsigned &)      { return  "%u"; }
template<> inline const char *FmtString(const unsigned long &) { return "%lu"; }
template<> inline const char *FmtString(const float &)         { return  "%f"; }
template<> inline const char *FmtString(const double &)        { return "%lf"; }
template<> inline const char *FmtString(const char &)          { return  "%c"; }


// Generate SQL-quoted version of string.  If EmptyIsNull is set, an empty
// string will generate the null value rather than an empty string.
template<typename T> std::string Quote(const T &Obj, bool EmptyIsNull=false);


// std::string version, on which the other versions are built
template<> inline std::string Quote(const std::string &Obj, bool EmptyIsNull)
{
    if (EmptyIsNull && Obj.empty())
        return "null";

    std::string Result = "'" + Obj + "'";

    for (std::string::size_type i = Result.size()-2; i > 0; --i)
        if ((Result[i] == '\'') ||
	    (Result[i] == '\\'))
	    Result.insert(i, 1, Result[i]);

    return Result;
}


// In the special case of const char *, the null pointer is represented as
// the null value.
template<> inline std::string Quote(const char *const & Obj, bool EmptyIsNull)
{
  if (!Obj) return "null";
  return Quote(std::string(Obj), EmptyIsNull);
}


// Generic version
template<typename T> inline std::string Quote(const T &Obj, bool EmptyIsNull)
{
  // TODO: Find a decent way to determine max string length at compile time
  char Buf[500];
  sprintf(Buf, FmtString(Obj), Obj);
  return Quote(std::string(Buf), EmptyIsNull);
}


}


#ifdef LACK_ITERATOR
// Ugly: substitute for real iterator struct if we didn't get one in <iterator>
namespace std
{
template<class Cat, class T, class Dist> struct iterator
{
  typedef Cat iterator_category;
  typedef T value_type;
  typedef Dist difference_type;
  typedef T *pointer;
  typedef T &reference;
};
}
#endif // LACK_ITERATOR


#endif

